use strict;
use Coro;
use Coro::Channel;
use Coro::Semaphore;
use Coro::Timer;
use URI;
use FurlX::Coro;
use Web::Query;
use Try::Tiny;

my @done;
my @fail;
sub done { push @done, $_[0] }
sub fail { push @fail, $_[0] }

my %queue;

sub queue {
    my $name = shift;
    $queue{$name} ||= Coro::Channel->new;
}

sub logger {
    my $msg = shift;
    warn localtime . sprintf ": %s %s\n", Coro::current->desc, $msg;
}

# 全体の同時接続数制限
my $global_lock = Coro::Semaphore->new(5);
sub global_lock {
    $global_lock->guard;
}

# ホスト毎の接続数制限
my %lock;
my $use_sleep = 1;
sub host_lock {
    my $url = shift;
    my $host = URI->new($url)->host;
    my $sem = $lock{$host} ||= Coro::Semaphore->new(1);
    # 最後の接続から一定時間waitを入れる
    if ($use_sleep && $sem->count == 0) {
        my $guard = $sem->guard;
        Coro::Timer::sleep 3;
        return $guard;
    }
    $sem->guard;
}

sub fetcher {  
    my $url = queue("fetch")->get;
    my $lock = host_lock($url);
    my $glock = global_lock();
    my $ua = FurlX::Coro->new;
    logger($url);
    my $res = $ua->get($url);
    queue("parse")->put([$url, $res]);
}

sub parser {
    my $data = queue("parse")->get;
    my ($url, $res) = @$data;
    logger($url);
    # タイトルを抜き出す場合
    my $title = Web::Query->new_from_html($res->content)->find("title")->text;
    queue("update")->put([$url, $title])
}

sub updater {
    my $data = queue("update")->get;
    my ($url, $res) = @$data;
    logger($url);
    warn $res;
    done($res);
}
sub create_worker {
    my ( $name, $code, $num ) = @_;
    for ( 0 .. $num ) {
        my $desc = $name . "_" . $_;
        async_pool {
            Coro::current->desc($desc);
            while (1) {
                try {
                    $code->() 
                } catch { 
                    warn $_;
                    fail($_);
                }
            }
        }
    }
}

create_worker( fetcher => \&fetcher, 1000 );
create_worker( parser  => \&parser,  1 );
create_worker( updater => \&updater, 1 );

my @list = qw(
    http://localhost/1
    http://localhost/2
    http://local.example.com/2
    http://local.example.com/3
);

my $stop_flag = 0;
my $force_exit = 0;
my $total = scalar @list;
my $main = Coro::current;

$SIG{INT} = sub {
    if ($stop_flag == 1) { $force_exit = 1 }
    $stop_flag = 1;
    if ($force_exit) { 
       die "exit";
    }
};

# manager
async {
    while (1) {
        Coro::Timer::sleep 1;
        # 作業途中のデータが失われないようにすると良い
        # 新規fetchのみ停止して残りの処理が終わるのを待つ, etc
        if ($stop_flag) {
            warn "signal recieved!!! ";
            async { 
                Coro::Timer::sleep 10;
                $stop_flag = 0
            }
        }
        warn sprintf "Task: %s/%s Fail: %s", scalar @done, $total, scalar @fail;
        my $done = scalar @done + scalar @fail;
        if ($done == $total) {
            warn "All task done!";
            $main->ready;
        }
    }
};

queue("fetch")->put($_) for @list;
schedule;
